import os

import pytest
from . import CodebergPackageDownloader, KnownPackages, PackageDescription


def run_tests_against_codeberg():
    if os.environ.get("HTTP_TESTS"):
        return False
    return True


@pytest.mark.skipif(run_tests_against_codeberg(), reason="Makes requests to codeberg")
def test_has_a_version():
    pd = PackageDescription(KnownPackages.connect, "0.1.3")
    downloader = CodebergPackageDownloader(pd)

    assert downloader.has_a_version("hollo", "0.6.19")
    assert not downloader.has_a_version("hollo", "0.6.20")


@pytest.mark.skipif(run_tests_against_codeberg(), reason="Makes requests to codeberg")
def test_no_results():
    pd = PackageDescription(KnownPackages.connect, "4520.0.0")
    downloader = CodebergPackageDownloader(pd)

    assert not downloader.has_a_version("hollo", "0.6.20")



from .types import FeatureResult


feature_info = {
    "name": "Can fetch public timeline",
    "status": "passed",
    "steps": [
        {
            "name": "Given A Fediverse application",
            "status": "passed",
            "start": 1766579158709,
            "stop": 1766579158709,
        },
        {
            "name": "When a message is posted from the application",
            "status": "passed",
            "start": 1766579158710,
            "stop": 1766579158814,
        },
        {
            "name": 'Then "pasture-one-actor" can read it',
            "status": "passed",
            "attachments": [
                {
                    "name": "Created object",
                    "source": "ebbc2ec6-179a-4061-87e6-495e8c06411f-attachment.json",
                    "type": "application/json",
                }
            ],
            "start": 1766579158814,
            "stop": 1766579158863,
        },
    ],
    "start": 1766562342505,
    "stop": 1766562343904,
    "uuid": "c8244d31-cfd8-489a-a76a-69737296e178",
    "historyId": "8aa6b52f61f474116aa15483f5f3bdf6",
    "testCaseId": "6acc62704f6f098cc2540004d0494fcc",
    "fullName": "mitra: Funfedi Connect support: Can fetch public timeline",
    "labels": [
        {"name": "severity", "value": "normal"},
        {"name": "tag", "value": "public-timeline"},
        {"name": "feature", "value": "mitra: Funfedi Connect support"},
        {"name": "framework", "value": "behave"},
        {"name": "language", "value": "cpython3"},
    ],
    "titlePath": ["features", "mitra: Funfedi Connect support"],
}


def test_feature_result():
    result = FeatureResult.from_data(feature_info)

    assert result.name == "mitra: Funfedi Connect support: Can fetch public timeline"
    assert result.status == "passed"
    assert result.tags == ["public-timeline"]
    assert result.attachments == {
        "Created object": "ebbc2ec6-179a-4061-87e6-495e8c06411f"
    }



from collections import defaultdict
from glob import glob
import re
import semver

pattern = re.compile(r"data/results_funfedi_connect/(.*)?/(.*)?_(.*).zip")


def to_semver(version, skip_prerelease=True):
    try:
        if version.count(".") == 1:
            version = "0." + version
        if "a" in version:
            version = version.split("a")[0]
        result = semver.Version.parse(version.removeprefix("v"))

        if result.prerelease and skip_prerelease:
            return semver.Version(0)
        return result
    except Exception:
        return semver.Version(0)


def later_version(a, b):
    """
    ```
    >>> later_version("v4.5.2", "v4.6.2")
    'v4.6.2'

    ```
    """
    if to_semver(a) > to_semver(b):
        return a
    return b


def latest_for_app_name(name: str):
    case_version = None
    app_version = None
    for x in glob(f"data/results_funfedi_connect/*/{name}_*"):
        m = re.match(pattern, x)
        if not m:
            raise Exception("Does not match regex")
        case_version = later_version(case_version, m.group(1))
        app_version = later_version(app_version, m.group(3))

    return case_version, app_version


def name_and_version(filename: str) -> tuple[str, str]:
    """
    ```
    >>> name_and_version("cattle_grid_0.5.20.zip")
    ('cattle_grid', '0.5.20')

    ```
    """
    split = filename.split("_")
    return "_".join(split[:-1]), split[-1].removesuffix(".zip")


def names_to_latest_dict(names: list[str]) -> dict[str, str]:
    """
    ```
    >>> names_to_latest_dict(["app_0.0.0"])
    {'app': '0.0.0'}

    >>> names_to_latest_dict(["app_0.0.0", "app_1.0.0"])
    {'app': '1.0.0'}

    ```
    """
    names_and_versions = [name_and_version(x) for x in names]

    name_to_version = defaultdict(list)

    for name, version in names_and_versions:
        name_to_version[name].append(version)

    return {
        name: sorted(versions, key=to_semver, reverse=True)[0]
        for name, versions in name_to_version.items()
    }



from dataclasses import dataclass
from glob import glob
from pathlib import Path
import zipfile

from funfedi_results.latest import name_and_version, names_to_latest_dict


@dataclass
class AllureResults:
    """Class to manage allure results"""

    path: Path = Path("allure-results")

    def clean(self):
        """Removes all files from the allure-results directory"""
        if self.path.exists():
            for child in self.path.iterdir():
                child.unlink()

        self.path.mkdir(exist_ok=True)

    @property
    def files_in_allure_dir(self):
        return glob(f"{self.path}/*.json")

    def create_zip(self, zipfilename: Path):
        """Zips all files in the allure-results directory into a file. Files
        are checked to have the naming convention `{app_name}_{app_version}.zip`"""
        name, version = name_and_version(zipfilename.name)
        if not (name and version):
            raise Exception(f"Invalid filename {zipfilename}")

        with zipfile.ZipFile(zipfilename, "w") as f:
            for name in self.files_in_allure_dir:
                new_name = name.split("/")[-1]
                f.write(name, new_name)

    def extract(self, file_directory: Path):
        """Takes all zip files from directory and extracts the latest
        version for each app into the allure-results directory"""
        self.clean()

        files = [x.name for x in file_directory.glob("*.zip")]
        latest = names_to_latest_dict(files)

        for name, version in latest.items():
            with zipfile.ZipFile(file_directory / f"{name}_{version}.zip") as f:
                f.extractall(self.path)



from dataclasses import dataclass
import os
import requests

from .package_description import PackageDescription, KnownPackages  # noqa: F401


class BadCodebergCredentialsException(Exception): ...


"""Thrown if CodebergUploader.from_env doesn't find the credentials"""


@dataclass
class CodebergUploader:
    """
    Helps uploading files to codeberg

    ```
    >>> uploader = CodebergUploader("user", "token")
    >>> uploader.upload_url(PackageDescription(KnownPackages.connect, "0.1.2"), "app_v1.2.3.zip")
    'https://user:token@codeberg.org/api/packages/funfedidev/generic/results_funfedi_connect/0.1.2/app_v1.2.3.zip'

    ```
    """

    user: str
    token: str

    @property
    def api_prefix(self) -> str:
        return f"https://{self.user}:{self.token}@codeberg.org/api/packages/funfedidev/generic/"

    def upload_url(self, package: PackageDescription, filename):
        return self.api_prefix + f"{package.path}/{filename}"

    def upload_file(self, package: PackageDescription, filename, file):
        """Uploads a file with url as in upload_url"""

        files = {"file": open(file, "rb")}
        url = self.upload_url(package, filename)
        result = requests.put(url, files=files)
        if result.status_code != 201:
            print(f"Upload to codeberg failed with status {result.status_code}")
            print(result.text)
        else:
            print("Uploaded to codeberg")

    @staticmethod
    def from_env():
        """Initializes uploader from CODERBERG_USER and CODEBERG_TOKEN environment variables"""
        user = os.environ.get("CODEBERG_USER")
        token = os.environ.get("CODEBERG_TOKEN")

        if not user or not token:
            raise BadCodebergCredentialsException(
                "Expected CODEBERG_USER and CODEBERG_TOKEN to be set"
            )

        return CodebergUploader(user=user, token=token)



from funfedi_results.archive.types import FeatureResult
from . import transform_feature_result


def test_transform_result():
    result = FeatureResult(
        name="app",
        status="passed",
        tags=["webfinger", "other"],
        start=12,
        attachments={},
    )

    transformed = transform_feature_result(result)

    assert transformed == {"webfinger": "passed"}



import os
import pytest


from funfedi_results.codeberg import KnownPackages, PackageDescription
from funfedi_connect import ImplementedFeature

from .per_tag import passing_counts
from .types import ResultDescription


@pytest.mark.skipif(not os.path.exists("data"), reason="No test data")
def test_per_tag():
    package = PackageDescription(KnownPackages.connect, "0.2.0")
    desc = ResultDescription(package, "pleroma", "v2.10.0")

    result = passing_counts(desc)

    assert result[ImplementedFeature.event].total == 0
    assert result[ImplementedFeature.public_timeline].total == 1
    assert result[ImplementedFeature.post].total == 1
    assert result[ImplementedFeature.webfinger].total == 7



import pytest
from .uploader import BadCodebergCredentialsException, CodebergUploader


def test_from_env(monkeypatch):
    monkeypatch.setenv("CODEBERG_USER", "user")
    monkeypatch.setenv("CODEBERG_TOKEN", "token")

    uploader = CodebergUploader.from_env()

    assert uploader.user == "user"
    assert uploader.token == "token"


def test_from_env_bad_credentials(monkeypatch):
    monkeypatch.setenv("CODEBERG_USER", "")
    monkeypatch.setenv("CODEBERG_TOKEN", "")

    with pytest.raises(BadCodebergCredentialsException):
        CodebergUploader.from_env()



import json
import logging
import zipfile
from .types import FeatureResult, LoadedResult, ResultDescription

logger = logging.getLogger(__name__)


def name_to_type(name: str) -> str:
    """
    ```
    >>> name_to_type("verify/allure-results/fa6bc2a7-5743-432b-9ad6-0ddcf14619e0-container.json")
    'container'

    ```
    """

    return name.split(".")[0].split("-")[-1]


def load_archive(desc: ResultDescription) -> LoadedResult:
    result = LoadedResult([], [], {})
    with zipfile.ZipFile(desc.filename) as f:
        for name in f.namelist():
            parsed = json.loads(f.read(name))
            match name_to_type(name):
                case "container":
                    result.containers.append(parsed)
                case "result":
                    result.results.append(parsed)
                case "attachment":
                    attachment_name = name.removesuffix("-attachment.json").split("/")[
                        -1
                    ]
                    result.attachments[attachment_name] = parsed
                case _:
                    logger.warning("Unknown name %s", name)
            ...
    return result


def features_from_result(result: LoadedResult) -> list[FeatureResult]:
    features: dict[str, FeatureResult] = {}
    for x in result.results:
        title_path = x.get("titlePath")
        if not title_path:
            continue
        if title_path[0] == "features":
            parsed = FeatureResult.from_data(x)
            if parsed and parsed.name in features:
                if features[parsed.name].start < parsed.start:
                    features[parsed.name] = parsed
            else:
                features[parsed.name] = parsed

    return list(features.values())



from dataclasses import dataclass, field

from funfedi_connect import ImplementedFeature

from . import features_for_result

from .types import ResultDescription


@dataclass
class PassingCount:
    feature: ImplementedFeature
    passed: int = field(default=0)
    failed: int = field(default=0)

    @property
    def total(self):
        return self.passed + self.failed

    def add_pass(self):
        self.passed = self.passed + 1

    def add_fail(self):
        self.failed = self.passed + 1

    def status(self):
        if self.total == 0:
            return ""
        if self.passed == 0:
            return "❌"
        if self.failed == 0:
            return "✅"

        return f"{self.passed}/{self.total}"


def passing_counts(result_description: ResultDescription):
    data = features_for_result(result_description)

    counts = {feature: PassingCount(feature) for feature in ImplementedFeature}

    for feature_result in data:
        for tag in feature_result.tags:
            try:
                feature = ImplementedFeature.from_tag(tag)
                if feature_result.status == "passed":
                    counts[feature].add_pass()
                else:
                    counts[feature].add_fail()
            except Exception:
                ...

    return counts



from funfedi_results.archive.types import FeatureResult

from .parse import features_from_result, load_archive

from .types import ResultDescription

__all__ = ["ResultDescription"]


def features_for_result(desc: ResultDescription) -> list[FeatureResult]:
    archive = load_archive(desc)

    return features_from_result(archive)



from funfedi_results.archive.types import LoadedResult
from .codeberg.package_description import KnownPackages, PackageDescription
from .transform import transform_feature_result
from .archive.parse import features_from_result, load_archive
from .archive.types import FeatureResult, ResultDescription
from .latest import latest_for_app_name

__all__ = ["feature_status", "FeatureResult", "LoadedResult"]


def load_latest(app_name: str) -> LoadedResult:
    case_version, app_version = latest_for_app_name(app_name)
    if not case_version or not app_version:
        raise Exception("Failed to find test data")
    package = PackageDescription(KnownPackages.connect, case_version)

    desc = ResultDescription(package, app_name, app_version)

    return load_archive(desc)


def feature_status(app_name: str):
    result = load_latest(app_name)
    features = features_from_result(result)

    feature_status_map = {}
    for feature in features:
        feature_status_map.update(transform_feature_result(feature))

    return feature_status_map



from dataclasses import dataclass

from funfedi_results.codeberg import PackageDescription, KnownPackages  # noqa: F401


@dataclass
class ResultDescription:
    """
    ```
    >>> package = PackageDescription(KnownPackages.connect, "0.1.2")
    >>> t = ResultDescription(package, "mitra", "v4.15.0")
    >>> t.filename
    'data/results_funfedi_connect/0.1.2/mitra_v4.15.0.zip'

    ```
    """

    package: PackageDescription
    application: str
    application_version: str

    @property
    def filename(self):
        return f"data/{self.package.path}/{self.application}_{self.application_version}.zip"


@dataclass
class LoadedResult:
    containers: list[dict]
    results: list[dict]
    attachments: dict[str, dict]


@dataclass
class FeatureResult:
    name: str
    status: str
    tags: list[str]
    start: int
    attachments: dict[str, str]

    @staticmethod
    def from_data(data: dict):
        name = data.get("fullName", "-- missing --")
        status = data.get("status", "-- missing --")
        tags = [
            x.get("value", "") for x in data.get("labels", []) if x.get("name") == "tag"
        ]
        attachments = {}
        for step in data.get("steps", []):
            for attachment in step.get("attachments", []):
                a = attachment.get("source")
                a_name = attachment.get("name")
                if a and a_name:
                    attachments[a_name] = a.removesuffix("-attachment.json")

        return FeatureResult(
            name, status, tags, start=data.get("start", 0), attachments=attachments
        )



import pytest
import os

from .parse import features_from_result, load_archive
from .types import ResultDescription


from funfedi_results.codeberg import KnownPackages, PackageDescription

package = PackageDescription(KnownPackages.connect, "0.1.3")


@pytest.mark.skipif(not os.path.exists("data"), reason="No test data")
def test_load():
    desc = ResultDescription(package, "mitra", "v4.15.0")

    result = load_archive(desc)

    assert result


@pytest.mark.skipif(not os.path.exists("data"), reason="No test data")
def test_features():
    desc = ResultDescription(package, "mitra", "v4.15.0")
    result = load_archive(desc)
    assert len(features_from_result(result)) > 0



import json

import click

from funfedi_results.codeberg import CodebergPackageDownloader, PackageDescription


@click.command
@click.option(
    "--package", default="results_funfedi_connect", help="The results to display"
)
@click.option("--package_version", default="0.1.3", help="The version to display")
@click.option("--json", "display_as_json", is_flag=True, default=False)
def latest(package: str, package_version: str, display_as_json):
    package_description = PackageDescription(package, package_version)
    downloader = CodebergPackageDownloader(package_description)

    if display_as_json:
        print(json.dumps(downloader.latest_version_per_app(), indent=2))
    else:
        for name, version in downloader.latest_version_per_app().items():
            print(f"{name + ':':<20}{version}")


if __name__ == "__main__":
    latest()



import os
import zipfile

import pytest
from . import AllureResults


@pytest.fixture
def sample_dir(tmp_path):
    sample_dir = tmp_path / "a"
    sample_dir.mkdir()
    return sample_dir


@pytest.fixture
def sample_file(sample_dir):
    filename = sample_dir / "file.json"
    with open(filename, "w") as f:
        f.write('{"key":"value"}')

    return filename


def test_clean(sample_file, sample_dir):
    ar = AllureResults(path=sample_dir)
    ar.clean()

    assert not os.path.exists(sample_file)


def test_create_zip_file(sample_file, sample_dir, tmp_path):
    ar = AllureResults(path=sample_dir)

    results_dir = tmp_path / "r"
    results_dir.mkdir(exist_ok=True)
    results_file = results_dir / "results_0.1.0.zip"

    ar.create_zip(results_file)

    assert os.path.exists(results_file)

    with zipfile.ZipFile(results_file) as f:
        assert "file.json" in f.namelist()


def test_create_zip_file_fails_for_incorrect_name(sample_dir, tmp_path):
    ar = AllureResults(path=sample_dir)

    results_dir = tmp_path / "r"
    results_dir.mkdir(exist_ok=True)
    results_file = results_dir / "results.zip"

    with pytest.raises(Exception):
        ar.create_zip(results_file)


def test_extract(sample_file, sample_dir, tmp_path):
    ar = AllureResults(path=sample_dir)

    results_dir = tmp_path / "r"
    results_dir.mkdir(exist_ok=True)
    results_file = results_dir / "results_0.1.0.zip"
    ar.create_zip(results_file)
    ar.clean()

    ar.extract(results_dir)

    assert os.path.exists(sample_file)



import pytest
import os
from . import feature_status


@pytest.mark.skipif(not os.path.exists("data"), reason="No test data")
def test_feature_status():
    result = feature_status("mitra")

    assert result == {
        "post": "passed",
        "public_timeline": "passed",
        "webfinger": "passed",
    }



from dataclasses import dataclass
from enum import StrEnum
import semver

from funfedi_results.latest import to_semver


class KnownPackages(StrEnum):
    """Predefined packages"""

    parsing = "results_funfedi_parsing"
    connect = "results_funfedi_connect"


@dataclass
class PackageDescription:
    """Describes a package"""

    name: str | KnownPackages
    version: str

    @property
    def path(self):
        return f"{self.name}/{self.version}"

    def __post_init__(self):
        if to_semver(self.version) == semver.Version(0):
            raise Exception("version should be a valid version")



from dataclasses import dataclass
from pathlib import Path

import requests
from ..latest import names_to_latest_dict

from .package_description import PackageDescription, KnownPackages
from .uploader import CodebergUploader, BadCodebergCredentialsException

__all__ = [
    "CodebergPackageDownloader",
    "CodebergUploader",
    "PackageDescription",
    "KnownPackages",
    "BadCodebergCredentialsException",
]


@dataclass
class CodebergPackageDownloader:
    """Enables downloading the latest version of a package"""

    package: PackageDescription
    owner: str = "funfedidev"

    api_prefix: str = "https://codeberg.org/api"

    @property
    def filelist_url(self):
        """
        ```
        >>> downloader = CodebergPackageDownloader(PackageDescription("results_funfedi_connect", "0.1.3"))
        >>> downloader.filelist_url
        'https://codeberg.org/api/v1/packages/funfedidev/generic/results_funfedi_connect/0.1.3/files'

        ```
        """
        return f"{self.api_prefix}/v1/packages/{self.owner}/generic/{self.package.path}/files"

    def download_link(self, app_name, app_version):
        """
        ```
        >>> downloader = CodebergPackageDownloader(PackageDescription("results_funfedi_connect", "0.1.3"))
        >>> downloader.download_link("cattle_grid", "0.5.20")
        'https://codeberg.org/api/packages/funfedidev/generic/results_funfedi_connect/0.1.3/cattle_grid_0.5.20.zip'

        ```
        """
        filename = f"{app_name}_{app_version}.zip"
        return f"{self.api_prefix}/packages/{self.owner}/generic/{self.package.path}/{filename}"

    def has_a_version(self, app_name, app_version) -> bool:
        """Returns True if a result file already exists for app_name and app_version"""
        filename = f"{app_name}_{app_version}.zip"
        return filename in self.fetch_filelist()

    def fetch_filelist(self) -> list[str]:
        result = requests.get(self.filelist_url)

        if result.status_code == 404:
            return []

        result.raise_for_status()
        data = result.json()
        return [x.get("name") for x in data]

    def latest_version_per_app(self):
        """Returns the latest versions"""
        names = self.fetch_filelist()
        return names_to_latest_dict(names)

    def target_directory(self, directory: Path):
        return directory / self.package.path

    def download_latest(self, directory: Path):
        """Downloads the latest zip files and deposits them in directory"""

        names_and_versions = self.latest_version_per_app()
        target_directory = self.target_directory(directory)
        target_directory.mkdir(exist_ok=True, parents=True)

        for name, version in names_and_versions.items():
            download_link = self.download_link(name, version)
            filename = target_directory / download_link.split("/")[-1]

            if filename.exists():
                continue

            print(f"downloading {download_link}")

            response = requests.get(download_link)
            if response.status_code != 200:
                raise Exception(
                    f"Got response code {response.status_code} for downloading {download_link}"
                )
            with open(filename, "wb") as f:
                f.write(response.content)



import click
from funfedi_connect import ImplementedFeature

from .archive.per_tag import passing_counts

from .archive import features_for_result
from .archive.types import ResultDescription

from .codeberg import CodebergPackageDownloader, PackageDescription


def display_long(package_description, latest):
    for app_name, app_version in latest.items():
        print(f"{app_name}: {app_version}")

        result_description = ResultDescription(
            package_description, app_name, app_version
        )
        features = features_for_result(result_description)

        for f in features:
            print(f"{f.name}: {f.status}")
        print()


def display_snippet(package_description, latest):
    result_string = "| Name | Version | "
    result_string += " | ".join(str(f) for f in ImplementedFeature) + " |"
    print(result_string)

    print("| " + " | ".join(["---"] * (2 + len(ImplementedFeature))) + " |")

    for app_name, app_version in latest.items():
        result_description = ResultDescription(
            package_description, app_name, app_version
        )
        counts = passing_counts(result_description)

        result_string = f"| {app_name} | {app_version} |"
        for feature in ImplementedFeature:
            result_string += f" {counts[feature].status()} |"

        print(result_string)


@click.command
@click.pass_context
@click.option(
    "--package", default="results_funfedi_connect", help="The results to display"
)
@click.option("--package_version", default="0.1.3", help="The version to display")
@click.option("--display")
def display_command(ctx: click.Context, package: str, package_version: str, display):
    package_description = PackageDescription(package, package_version)
    downloader = CodebergPackageDownloader(package_description)
    downloader.download_latest(ctx.obj["data_dir"])

    latest = downloader.latest_version_per_app()

    if display == "snippet":
        display_snippet(package_description, latest)
    else:
        display_long(package_description, latest)


if __name__ == "__main__":
    display_command()



from funfedi_connect import ImplementedFeature
from funfedi_results.archive.types import FeatureResult


def tags_to_features(tags: list[str]) -> list[str]:
    """
    ```
    >>> tags_to_features(["webfinger", "other"])
    ['webfinger']

    ```
    """
    result = []
    for tag in tags:
        try:
            result.append(str(ImplementedFeature.from_tag(tag)))
        except Exception:
            ...
    return result


def transform_feature_result(result: FeatureResult):
    return {feature: result.status for feature in tags_to_features(result.tags)}



import pytest
from .package_description import KnownPackages, PackageDescription


def test_package_description():
    PackageDescription("name", "0.1.2")


def test_package_description_failure():
    with pytest.raises(Exception):
        PackageDescription("name", "xxx")


def test_package_description_path():
    p = PackageDescription(KnownPackages.parsing, "0.1.4")

    assert p.path == "results_funfedi_parsing/0.1.4"



from pathlib import Path
import click

from .latest.__main__ import latest

from .allure import AllureResults
from .display import display_command
from .codeberg import CodebergPackageDownloader, PackageDescription


@click.group
@click.option("--data_dir", default="./data")
@click.pass_context
def main(ctx: click.Context, data_dir):
    ctx.ensure_object(dict)
    ctx.obj["data_dir"] = Path(data_dir)
    ctx.obj["data_dir"].mkdir(exist_ok=True)


@main.command
@click.option(
    "--package", default="results_funfedi_connect", help="The results to display"
)
@click.option("--package_version", default="0.1.3", help="The version to display")
@click.option(
    "--as_allure_results",
    is_flag=True,
    default=False,
    help="Extracts into the allure-results directory",
)
@click.pass_context
def download(
    ctx: click.Context, package: str, package_version: str, as_allure_results: bool
):
    package_description = PackageDescription(package, package_version)
    downloader = CodebergPackageDownloader(package_description)
    downloader.download_latest(ctx.obj["data_dir"])

    if as_allure_results:
        allure_results = AllureResults()
        file_directory = downloader.target_directory(ctx.obj["data_dir"])
        allure_results.extract(file_directory)


main.add_command(display_command, "display")
main.add_command(latest, "latest")

if __name__ == "__main__":
    main()
